# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2017-03-31 17:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0007_task_is_persistent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='is_persistent',
            field=models.BooleanField(default=True, verbose_name='\u0412\u044b\u043f\u043e\u043b\u043d\u044f\u0442\u044c \u043f\u043e\u043a\u0430 \u043d\u0435 \u0437\u0430\u0432\u0435\u0440\u0448\u0438\u0442\u0441\u044f \u0443\u0441\u043f\u0435\u0448\u043d\u043e'),
        ),
    ]



# -*- coding: utf-8 -*-
import importlib
import logging
import os
import sys
from datetime import datetime, timedelta
from multiprocessing import Queue, Process
from time import sleep, time

import six
from django.db import connections
from django.db.models import Q
from django.utils import timezone

from botmanager import settings
from botmanager.management.commands import BotManagerBaseCommand, BotManagerBaseCommandException
from botmanager.models import Task
from botmanager.utils import management_lock

if six.PY2:
    from Queue import Empty as QueueEmpty
else:
    from queue import Empty as QueueEmpty

try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = lambda x: x


class BotManagerCommandException(BotManagerBaseCommandException):
    pass


class Command(BotManagerBaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('action', nargs='?', choices=('start', 'stop', 'restart'), default='start')
        parser.add_argument('--soft', action='store_true', default=False)
        parser.add_argument('--without_sheduller', action='store_true', default=False)
        parser.add_argument('--task_cls', type=str)

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)

        self.config = settings.MAIN_CONFIG

        if 'tasks' not in self.config:
            raise BotManagerCommandException(u"settings.BOTMANAGER_CONFIG['tasks'] not found")

        if len(self.config['tasks']) == 0:
            raise BotManagerCommandException(u"settings.BOTMANAGER_CONFIG['tasks'] is empty")

    @staticmethod
    def import_from_string(val):
        try:
            parts = val.split('.')
            module_path, class_name = '.'.join(parts[:-1]), parts[-1]
            module = importlib.import_module(module_path)
            return getattr(module, class_name)
        except ImportError as e:
            raise ImportError(
                u"Could not import '%s' for setting. %s: %s." % (val, e.__class__.__name__, e)
            )

    @staticmethod
    def start_service(service):
        logging.info(u"Start task manager {0}".format(service.process_name))
        connections.close_all()
        try:
            service.run()
        except KeyboardInterrupt:
            pass
        finally:
            connections.close_all()

    def handle(self, *args, **options):
        if options['action'] in ['stop', 'restart']:
            if options['soft']:
                raise NotImplementedError("Soft stopping is not implemented")
            else:
                logging.info("BotManager stopped.")
                if options['action'] == 'stop':
                    return
        logging.info(u"Try start BotManager...")
        sleep(1)
        self.run(*args, **options)

    def __clean_dir(self, dir_path, life_hours, full_clean=False):
        if os.path.isdir(dir_path):
            files = os.listdir(dir_path)
            for f in files:
                log_file_path = os.path.join(dir_path, f)
                if os.path.isfile(os.path.join(dir_path, f)):
                    hours = (int(time()) - int(os.path.getctime(log_file_path))) / 3600
                    if hours >= life_hours or full_clean:
                        logging.debug('clean file: {0} ({1} days)'.format(log_file_path, hours))
                        os.remove(log_file_path)

                elif os.path.isdir(log_file_path):
                    self.__clean_dir(log_file_path, life_hours, full_clean=full_clean)

    def remove_logs(self, logs_life_hours):
        log_config = self.config['logs']
        logs_dir = log_config['dir'] if 'dir' in log_config else None

        if logs_dir:
            logging.info("Start clean logs dir {0} with logs_life_hours={1}".format(
                logs_dir, logs_life_hours
            ))
            task_types = []
            for dir_name in os.listdir(logs_dir):
                dir_path = os.path.join(logs_dir, dir_name)
                if os.path.isdir(dir_path):
                    task_types.append(dir_name)

            for task_type in task_types:
                task_dir = os.path.join(logs_dir, task_type)
                self.__clean_dir(task_dir, life_hours=logs_life_hours)

    def remove_tasks(self, tasks_life_hours):
        Task.objects.filter(
            Q(parent=None) | Q(parent__is_complete=True),
            is_complete=True,
            finish_dt__lt=timezone.now() - timedelta(hours=tasks_life_hours)
        ).delete()

    @management_lock
    def run(self, *args, **options):

        self.set_logging()

        logging.info(u"Starting BotManager")
        setproctitle(u"BotManager.General")

        current_pid = os.getpid()

        processes = []
        queue_dict = {}
        for task_class_string, processes_count in self.config['tasks'].items():
            cls_name = task_class_string.split('.')[-1]
            if options['task_cls'] and cls_name != options['task_cls']:
                logging.info('Continue task class {}'.format(cls_name))
                continue
            task_class = Command.import_from_string(task_class_string)
            self._validate_task_class(task_class)
            # task_class.prepare()

            maxsize = processes_count * 10
            queue_dict[task_class.name] = Queue(maxsize=maxsize)
            queue_dict[task_class.name].maxsize = maxsize

            for i in range(processes_count):
                tm = TaskManager(task_class, queue_dict[task_class.name], i + 1, current_pid)
                p = Process(target=Command.start_service, args=(tm,))
                p.name = tm.process_name
                p.daemon = True
                p.start()
                processes.append(p)

        tf = TaskFetcher(queue_dict, current_pid)
        p = Process(target=Command.start_service, args=(tf,))
        p.name = tf.process_name
        p.daemon = True
        p.start()
        processes.append(p)

        if not options.get('without_sheduller'):
            ts = TaskSheduler(self.config, current_pid)
            p = Process(target=Command.start_service, args=(ts,))
            p.name = ts.process_name
            p.daemon = True
            p.start()
            processes.append(p)

        next_clean_logs_dt = next_clean_tasks_dt = datetime.now() + timedelta(seconds=10)
        logs_life_hours = self.config['logs']['logs_life_hours']
        tasks_life_hours = self.config['logs']['success_tasks_life_hours']
        need_remove_logs = self.config['logs']['task_logs_separated'] and logs_life_hours is not None
        need_remove_tasks = tasks_life_hours is not None

        logging.info("logs_life_hours={0}, tasks_life_hours={1}, need_remove_logs={2}, need_remove_tasks={3}".format(
            logs_life_hours, tasks_life_hours, need_remove_logs, need_remove_tasks
        ))

        while True:
            statuses = set(map(lambda x: x.is_alive(), processes))
            if len(statuses) == 1 and list(statuses)[0] is False:
                break
            else:
                if need_remove_logs:
                    if next_clean_logs_dt <= datetime.now():
                        logging.info("Start delete old separated tasks logs...")
                        self.remove_logs(logs_life_hours)
                        next_clean_logs_dt += timedelta(hours=1)
                        logging.info("Delete old logs finished, next at {0}".format(next_clean_logs_dt))

                if need_remove_tasks:
                    if next_clean_tasks_dt <= datetime.now():
                        logging.info("Start delete old success tasks...")
                        self.remove_tasks(tasks_life_hours)
                        next_clean_tasks_dt += timedelta(hours=1)
                        logging.info("Delete old success tasks finished, next at {0}".format(next_clean_tasks_dt))
            sleep(5)

        logging.info(u"BotManager is stopped.")

    def _validate_task_class(self, _class):
        if not hasattr(_class, 'actions'):
            raise BotManagerCommandException('No actions in class {}'.format(_class.__name__))

        if not _class.actions:
            raise BotManagerCommandException('Empty actions in class {}'.format(_class.__name__))

        for action_str in _class.actions:
            if not hasattr(_class, action_str):
                raise BotManagerCommandException('Invalid action {} in class {}'.format(
                    action_str, _class.__name__
                ))


class TaskSheduler(object):
    SET_PERIOD_SECONDS = 5

    def __init__(self, config, parent_pid):
        self.process_name = "BotManager.TaskSheduler"
        self.config = config
        self.parent_pid = parent_pid
        self.shedule_cache = {}

    def run(self):
        setproctitle(self.process_name)
        while True:

            if os.getppid() != self.parent_pid:
                logging.info(u"Parent process is die. Exit..")
                break

            for task_class_string, processes_count in self.config['tasks'].items():
                task_class = Command.import_from_string(task_class_string)

                if self._time_to_set_tasks_for(task_class) and task_class.SELF_SUPPORT:
                    self.shedule_cache[task_class.name] = datetime.now()
                    try:
                        task_class.set_tasks()
                    except Exception as e:
                        logging.exception(e)

            sleep(self.SET_PERIOD_SECONDS)

    def _time_to_set_tasks_for(self, task_class):
        last_shedule_time = self.shedule_cache.get(task_class.name)
        shedule_period = getattr(task_class, 'SHEDULE_PERIOD_SECONDS', self.SET_PERIOD_SECONDS)
        return not last_shedule_time or datetime.now() > (last_shedule_time + timedelta(seconds=shedule_period))


class TaskFetcher(object):
    FETCH_PERIOD_SECONDS = 5

    def __init__(self, queue_dict, parent_pid):
        self.process_name = "BotManager.TaskFetcher"
        self.queue_dict = queue_dict
        self.parent_pid = parent_pid

    def run(self):
        """
        Сделано на скорую руку. На текущих объемах этого более чем достаточно, но при увеличении кол-ва тасков,
        а так же типов тасков, надо решить следующие задачи:

        1. Запрос к БД берет все невыполненные таски хотя размеры очередей ограничены,
           значит брать нужно лимитированное кол-во каждого типа таска (столько сколько может влезть).
        2. Если брать задачи каждого типа таска отдельно, чтоб учитывать лимиты из п1, то когда типов тасков будет
           много, например 100. то наполнение очереди будет отставать.
        3. Учитывать queue_key для группировки
        4. в данный момент берутся из БД заведомо лишние задачи, у которых can_execute==False
        5. in_process ставится отдельно на каждую взятую задачу а не разово на всех.
        """
        setproctitle(self.process_name)
        Task.objects.filter(in_process=True).update(in_process=False)
        while True:
            try:

                if os.getppid() != self.parent_pid:
                    logging.info(u"Parent process is die. Exit..")
                    break

                tasks = Task.objects.filter(
                    is_complete=False
                ).order_by('-priority', 'id')

                running_tasks = {}
                for task in tasks:
                    logging.info('Fetching task {}'.format(task))

                    if task.in_process:
                        logging.info('Task in process. Putting to running_tasks dict {} {}'.format(task, task.queue_key))
                        running_tasks[task.queue_key] = task.id
                        continue
                    else:
                        if task.queue_key in running_tasks:
                            logging.info(
                                u"Do not put to queue task {0}, because already queued "
                                u"task {1} with same queue_key".format(
                                    task.id, running_tasks[task.queue_key]
                                )
                            )
                            continue

                    if task.can_execute:
                        if task.name not in self.queue_dict:
                            logging.info('Task can execute but there is no queue_dict for task name {} {} {}. Passing...'.format(
                                task.name, task, task.queue_key
                            ))
                            pass
                            # logging.warning(U"Undefined task type {0} with id {1}".format(task.name, task.id))
                        else:
                            if self._get_qsize(task) < self.queue_dict[task.name].maxsize:
                                logging.info('Task can execute. Putting to queue_dict with key {} {}'.format(
                                    task.name, task
                                ))
                                task.in_process = True
                                task.save(update_fields=('in_process',))
                                running_tasks[task.queue_key] = task.id
                                self.queue_dict[task.name].put(task)
            except KeyboardInterrupt:
                sys.exit()
            except Exception as e:
                logging.exception(e)

            sleep(self.FETCH_PERIOD_SECONDS)

    def _get_qsize(self, task):
        try:
            return self.queue_dict[task.name].qsize()
        except NotImplementedError:
            # Queue.qsize not working on Mac OS X
            from random import randint
            return randint(0, self.queue_dict[task.name].maxsize)


class TaskManager(object):
    def __init__(self, task_class, queue, process_num, parent_pid):
        self.task_class = task_class
        self.queue = queue
        self.process_num = process_num
        self.process_name = u"BotManager.{0}.{1}".format(self.task_class.name, self.process_num)
        self.parent_pid = parent_pid

    def __str__(self):
        return self.process_name

    def run_task(self, task):
        logging.info(u"Start {0}".format(task))
        self.task_class(task).run()

    def run(self):
        setproctitle(self.process_name)
        while True:
            try:

                if os.getppid() != self.parent_pid:
                    logging.info(u"Parent process is die. Exit..")
                    break

                task = self.queue.get_nowait()
                try:
                    self.run_task(task)
                except Exception as e:
                    logging.exception(u"Worker {0} catch exception on task {1}: {2}".format(
                        self, task.id, e
                    ))
                finally:
                    pass
            except QueueEmpty:
                sleep(1)
            except Exception as e:
                logging.exception(
                    u"Error in queue preparing: %s".format(e)
                )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2017-02-01 18:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0004_auto_20161219_1931'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child_tasks', to='botmanager.Task', verbose_name='\u0420\u043e\u0434\u0438\u0442\u0435\u043b\u044c\u0441\u043a\u0430\u044f \u0437\u0430\u0434\u0430\u0447\u0430'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2017-05-31 13:21
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0009_task_extra_params'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='failed_action',
            field=models.CharField(blank=True, max_length=256, null=True, verbose_name='Action, \u043d\u0430 \u043a\u043e\u0442\u043e\u0440\u043e\u043c \u043f\u0440\u043e\u0438\u0437\u043e\u0448\u043b\u0430 \u043e\u0448\u0438\u0431\u043a\u0430'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-12-08 15:29
from __future__ import unicode_literals

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0002_auto_20161208_1406'),
    ]

    operations = [
        migrations.RenameField(
            model_name='task',
            old_name='retry_count',
            new_name='attempt_count',
        ),
        migrations.RemoveField(
            model_name='task',
            name='last_retry_dt',
        ),
        migrations.RemoveField(
            model_name='task',
            name='max_retry_count',
        ),
        migrations.RemoveField(
            model_name='task',
            name='retry_period',
        ),
        migrations.AddField(
            model_name='task',
            name='attempt_period',
            field=models.DurationField(default=datetime.timedelta(0, 3600), verbose_name='\u041f\u0435\u0440\u0438\u043e\u0434 \u043c\u0435\u0436\u0434\u0443 \u043f\u043e\u043f\u044b\u0442\u043a\u0430\u043c\u0438'),
        ),
        migrations.AddField(
            model_name='task',
            name='last_attempt_dt',
            field=models.DateTimeField(null=True, verbose_name='\u041f\u043e\u0441\u043b\u0435\u0434\u043d\u044f\u044f \u043f\u043e\u043f\u044b\u0442\u043a\u0430 \u0432\u044b\u043f\u043e\u043b\u043d\u0435\u043d\u0438\u044f'),
        ),
        migrations.AddField(
            model_name='task',
            name='max_attempt_count',
            field=models.IntegerField(null=True, verbose_name='\u041c\u0430\u043a\u0441\u0438\u043c\u0430\u043b\u044c\u043d\u043e\u0435 \u043a\u043e\u043b-\u0432\u043e \u043f\u043e\u043f\u044b\u0442\u043e\u043a \u0432\u044b\u043f\u043e\u043b\u043d\u0435\u043d\u0438\u0435 (\u0435\u0441\u043b\u0438 None, \u0442\u043e \u0431\u0435\u0441\u043a\u043e\u043d\u0435\u0447\u043d\u043e)'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-12-19 19:31
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0003_auto_20161208_1529'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='create_dt',
            field=models.DateTimeField(auto_now_add=True, verbose_name='\u0414\u0430\u0442\u0430 \u0441\u043e\u0437\u0434\u0430\u043d\u0438\u044f \u0437\u0430\u0434\u0430\u0447\u0438'),
        ),
        migrations.AlterField(
            model_name='task',
            name='failed_action',
            field=models.CharField(blank=True, max_length=32, null=True, verbose_name='Action, \u043d\u0430 \u043a\u043e\u0442\u043e\u0440\u043e\u043c \u043f\u0440\u043e\u0438\u0437\u043e\u0448\u043b\u0430 \u043e\u0448\u0438\u0431\u043a\u0430'),
        ),
        migrations.AlterField(
            model_name='task',
            name='finish_dt',
            field=models.DateTimeField(blank=True, null=True, verbose_name='\u0414\u0430\u0442\u0430 \u0437\u0430\u0432\u0435\u0440\u0448\u0435\u043d\u0438\u044f'),
        ),
        migrations.AlterField(
            model_name='task',
            name='input',
            field=jsonfield.fields.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='last_attempt_dt',
            field=models.DateTimeField(blank=True, null=True, verbose_name='\u041f\u043e\u0441\u043b\u0435\u0434\u043d\u044f\u044f \u043f\u043e\u043f\u044b\u0442\u043a\u0430 \u0432\u044b\u043f\u043e\u043b\u043d\u0435\u043d\u0438\u044f'),
        ),
        migrations.AlterField(
            model_name='task',
            name='last_error',
            field=jsonfield.fields.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='last_error_dt',
            field=models.DateTimeField(blank=True, null=True, verbose_name='\u0414\u0430\u0442\u0430 \u043f\u043e\u0441\u043b\u0435\u0434\u043d\u0435\u0433\u043e fail-\u0430'),
        ),
        migrations.AlterField(
            model_name='task',
            name='max_attempt_count',
            field=models.IntegerField(blank=True, null=True, verbose_name='\u041c\u0430\u043a\u0441\u0438\u043c\u0430\u043b\u044c\u043d\u043e\u0435 \u043a\u043e\u043b-\u0432\u043e \u043f\u043e\u043f\u044b\u0442\u043e\u043a \u0432\u044b\u043f\u043e\u043b\u043d\u0435\u043d\u0438\u0435 (\u0435\u0441\u043b\u0438 None, \u0442\u043e \u0431\u0435\u0441\u043a\u043e\u043d\u0435\u0447\u043d\u043e)'),
        ),
        migrations.AlterField(
            model_name='task',
            name='output',
            field=jsonfield.fields.JSONField(blank=True, null=True),
        ),
    ]



import fcntl
import logging
import os
def management_lock(view_func):

    def wrapper_lock(*args, **kwargs):
        try:
            lock_file_path = "{0}.lock".format(args[0].__class__.__module__.split('.')[-1])
            f = open(lock_file_path, 'w')
            fcntl.lockf(f, fcntl.LOCK_EX + fcntl.LOCK_NB)
        except IOError:
            logging.debug("Process already is running.")
            os._exit(1)
        return view_func(*args, **kwargs)

    wrapper_lock.view_func = view_func.view_func if hasattr(view_func, 'view_func') else view_func
    return wrapper_lock



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2017-04-25 16:51
from __future__ import unicode_literals

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0008_auto_20170331_1752'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='extra_params',
            field=jsonfield.fields.JSONField(blank=True, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from django.contrib import admin
from django.utils.html import escape
from botmanager.models import Task


class TaskAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'is_complete', 'in_process', 'is_failed',
                    'create_dt', 'finish_dt', 'error_field', 'attempt_period', 'attempt_count', 'input_field']
    list_filter = ('is_complete', 'is_failed', 'in_process', 'name')
    search_fields = ('id', 'name')

    def get_queryset(self, request):
        qs = super(TaskAdmin, self).get_queryset(request)
        return qs

    def error_field(self, obj):
        err_max_len = 50
        if isinstance(obj.last_error, dict) and 'error' in obj.last_error:
            return "Failed dt: {0}<br />Failed action: {1}<br />Error: {2}".format(
                obj.last_error_dt.strftime("%Y-%m-%d %H:%M:%S"),
                obj.failed_action,
                '<span style="cursor:help" title="{0}">{1} ...</span>'.format(
                    escape(obj.last_error['error']),
                    escape(obj.last_error['error'][0:err_max_len])
                ) if len(obj.last_error['error']) > err_max_len else obj.last_error['error']
            )
        else:
            return None

    error_field.short_description = u"Ошибка"
    error_field.allow_tags = True

    def input_field(self, obj):
        return obj.input

    input_field.short_description = "Вводные данные"

admin.site.register(Task, TaskAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.db import models
from jsonfield import JSONField
from datetime import timedelta, datetime
from django.utils import timezone


class Task(models.Model):

    STATUS_NEW = 'NEW'
    STATUS_IN_PROCESS = 'IN_PROCESS'
    STATUS_OK = 'OK'

    STATUSES = (
        (STATUS_NEW, u'В ожидании'),
        (STATUS_IN_PROCESS, u'Идет загрузка'),
        (STATUS_OK, u'Данные загружены'),
    )

    name = models.CharField(max_length=256, verbose_name=u'Название')
    create_dt = models.DateTimeField(verbose_name=u"Дата создания задачи", auto_now_add=True)
    finish_dt = models.DateTimeField(verbose_name=u"Дата завершения", null=True, blank=True)
    is_complete = models.BooleanField(verbose_name=u"Выполнена", default=False)
    is_failed = models.BooleanField(verbose_name=u"Аварийное завершение", default=False)
    is_persistent = models.BooleanField(verbose_name=u"Выполнять пока не завершится успешно", default=True)
    in_process = models.BooleanField(verbose_name=u"В процессе", default=False)

    input = JSONField(null=True, blank=True)
    output = JSONField(null=True, blank=True)
    last_error = JSONField(null=True, blank=True)
    extra_params = JSONField(null=True, blank=True)

    last_error_dt = models.DateTimeField(verbose_name=u"Дата последнего fail-а", null=True, blank=True)
    queue_key = models.CharField(max_length=32, verbose_name=u'Ключ группировки задач')
    failed_action = models.CharField(max_length=256, verbose_name=u'Action, на котором произошла ошибка', null=True, blank=True)
    attempt_count = models.SmallIntegerField(verbose_name=u'Счетчик попыток', default=0)
    last_attempt_dt = models.DateTimeField(verbose_name=u"Последняя попытка выполнения", null=True, blank=True)
    max_attempt_count = models.IntegerField(
        verbose_name=u"Максимальное кол-во попыток выполнение (если None, то бесконечно)", null=True, blank=True
    )
    attempt_period = models.DurationField(verbose_name=u'Период между попытками', default=timedelta(hours=1))
    parent = models.ForeignKey('self', verbose_name=u'Родительская задача', null=True, blank=True, related_name='child_tasks', on_delete=models.CASCADE)
    priority = models.SmallIntegerField(default=0)

    def __unicode__(self):
        return "Task {0} #{1}".format(self.name, self.id)

    @property
    def can_execute(self):
        if not self.is_complete and not self.in_process:
            if self.is_failed:
                if self.last_attempt_dt + self.attempt_period > timezone.now():
                    return False
                else:
                    if self.max_attempt_count is None:
                        return True
                    else:
                        if self.attempt_count < self.max_attempt_count:
                            return True
                        else:
                            return False  # надо пофиксить is_complete ?
            else:
                return True
        else:
            return False

    class Meta:
        db_table = 'botmanager_task'
        verbose_name = u"Задача для BotManager"
        verbose_name_plural = u"Задачи для BotManager"


from django.core.management.base import BaseCommand
from botmanager import settings
import logging
import logging.config


class BotManagerBaseCommandException(Exception):
    pass


class BotManagerBaseCommand(BaseCommand):

    def set_logging(self):
        if settings.TASKS_LOGGING is not None:
            logging.config.dictConfig(settings.TASKS_LOGGING)
            root_logger = logging.getLogger()
            root_logger.handlers = logging.getLogger('default').handlers
            root_logger.setLevel(settings.DEFAULT_LOG_LEVEL)



# -*- coding: utf-8 -*-
import logging
import logging.config
import os
import traceback
import json
from botmanager import settings
from botmanager.models import Task
from django.conf import settings as project_settings
from django.db import connection
from datetime import datetime, timedelta
from hashlib import md5
from botmanager import BotManagerTaskError
from django.utils import timezone


LOGGING = settings.TASKS_LOGGING


class StopTaskSuccess(BotManagerTaskError):
    pass


class BotManagerBaseTask(object):

    actions = []
    name = 'unknown'
    DEFAULT_ATTEMPT_PERIOD = 1
    SELF_SUPPORT = True

    def __init__(self, task):
        self.task = task
        self.config = settings.MAIN_CONFIG
        if self.config['logs']['task_logs_separated']:
            self.set_task_logger()
        self._on_init()

    def get_log_file_name(self):
        return u"task{0}".format(self.task.id)

    @classmethod
    def prepare(cls):
        pass

    def _on_init(self):
        pass

    @classmethod
    def get_queue_key(cls, task_input):
        return md5('{}_{}'.format(cls.name, json.dumps(task_input)).encode('utf-8')).hexdigest()

    @classmethod
    def create(cls, **kwargs):
        attempt_period = kwargs.pop('attempt_period', timedelta(hours=cls.DEFAULT_ATTEMPT_PERIOD))
        is_unique = kwargs.pop('is_unique', False)

        main_params = kwargs.copy()
        main_params.update({
            'name': cls.name,
            'queue_key': cls.get_queue_key(kwargs.get('input')),
        })
        if is_unique:
            task, created = Task.objects.get_or_create(
                defaults={
                    'attempt_period': attempt_period,
                },
                is_complete=False,
                **main_params
            )
        else:
            task = Task.objects.create(
                attempt_period=attempt_period,
                **main_params
            )
            created = True

        if created:
            logging.info('Task created. Name {}. id {}. args {}'.format(cls.name, task.pk, task.input))
        else:
            logging.info('Task already exists and not finished. Name {}. id {}. args {}'.format(
                cls.name, task.pk, task.input
            ))

        return task, created

    def set_task_logger(self):
        level = self.config['logs']['level'] if 'level' in self.config['logs'] else settings.DEFAULT_LOG_LEVEL

        task_file_dir = os.path.join(self.config['logs']['dir'], self.name)
        task_file_path = os.path.join(task_file_dir, u'{0}.log'.format(self.get_log_file_name()))
        log_handler_name = 'botmanager.{0}'.format(self.name)

        if not os.path.exists(task_file_dir):
            os.mkdir(task_file_dir)

        # create task logger
        LOGGING.setdefault('handlers', {})
        LOGGING['handlers'][self.name] = {
            'level': level,
            'class': 'logging.FileHandler',
            'filename': task_file_path,
            'formatter': 'default',
            'encoding': 'utf-8',
        }
        LOGGING.setdefault('loggers', {})
        handlers = ['common_errors', 'common', self.name]

        if project_settings.DEBUG:
            handlers.append('console')

        LOGGING['loggers'][log_handler_name] = {
            'handlers': handlers,
            'level': 'INFO',
            'propagate': False
        }

        logging.config.dictConfig(LOGGING)

        # copy task logger handlers to root logger
        root_logger = logging.getLogger()
        root_logger.handlers = logging.getLogger(log_handler_name).handlers
        root_logger.setLevel(logging.INFO)

    def _set_output(self, data):
        self.task.output = data
        self.task.save(update_fields=('output',))

    @classmethod
    def set_tasks(cls, **kwargs):
        pass

    def _handle_error(self, e):
        pass

    def _on_finish(self, is_success=True, is_complete=True):
        pass

    def _log_task_finish(self, is_success):
        logging.info('Task {0} FINISHED: success={1}, now={2}, args {3}'.format(
            self.task.pk, is_success, datetime.now(), self.task.input
        ))

    def _on_start(self):
        pass

    def is_valid_input_or_success(self):
        u""" Если в задачу пришли невалидные параметры - задача считается успешно выполненой """
        return True

    def should_skip_action(self, action):
        pass

    def run(self):
        logging.info('Run task {} {}'.format(self.name, self.task.pk))
        self._on_start()

        if self.task.input is None:
            self.task.input = {}
        if self.task.output is None:
            self.task.output = {}

        is_failed = False
        is_error_handled = False

        if self.is_valid_input_or_success():
            for action in self.actions:

                if self.should_skip_action(action):
                    logging.info('Skipping action because we should {}'.format(action))
                    continue

                if self.task.failed_action and self.task.failed_action != action:
                    logging.info('Task has failed action {}. ID {}. Skipping {}'.format(
                        self.task.failed_action, self.task.pk, action
                    ))
                    continue

                self.task.failed_action = None

                logging.info(u"Start action {}. id {}".format(action, self.task.pk))
                action_method = getattr(self, action)
                try:
                    action_method()
                    logging.info(u'End action {}. id {}'.format(action, self.task.pk))
                except StopTaskSuccess:
                    logging.info(u'Stop actions because StopTaskSuccess caught')
                    break
                except Exception as e:
                    logging.exception(e)
                    self.task.failed_action = action
                    is_error_handled = self._handle_error(e)
                    self.task.last_error_dt = timezone.now()
                    self.task.last_error = {
                        'error': str(e),
                        'trace': traceback.format_exc(),
                    }
                    logging.info('Exception caught and handled {}'.format(str(e)))
                    is_failed = not is_error_handled
                    break
        else:
            logging.warning(u'Task input is invalid and task marker as success.')

        self.task.is_failed = is_failed

        self.task.attempt_count += 1
        if is_failed and self.task.is_persistent:
            if self.task.max_attempt_count is None:
                is_complete = False
            else:
                if self.task.attempt_count < self.task.max_attempt_count:
                    is_complete = False
                else:
                    is_complete = True
        else:
            is_complete = True

        now = timezone.now()

        self.task.is_complete = is_complete
        if is_complete:
            self.task.finish_dt = now

        logging.info('Set task in_process False. id {}'.format(self.task.pk))
        self.task.in_process = False
        self.task.last_attempt_dt = now

        is_success = not is_failed or is_error_handled
        self._on_finish(is_success=is_success, is_complete=is_complete)
        self.task.save()
        logging.info('Closing connection to DB...')
        connection.close()
        logging.info('End run task {} id {}'.format(self.name, self.task.pk))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-12-08 14:06
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='last_error',
            field=jsonfield.fields.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='last_error_dt',
            field=models.DateTimeField(null=True, verbose_name='\u0414\u0430\u0442\u0430 \u043f\u043e\u0441\u043b\u0435\u0434\u043d\u0435\u0433\u043e fail-\u0430'),
        ),
    ]



from django.conf import settings
import sys
import os

MAIN_CONFIG = getattr(settings, 'BOTMANAGER_CONFIG', {})
if 'logs' not in MAIN_CONFIG:
    MAIN_CONFIG['logs'] = {}
if 'tasks' not in MAIN_CONFIG:
    MAIN_CONFIG['tasks'] = []

if 'task_logs_separated' not in MAIN_CONFIG['logs']:
    MAIN_CONFIG['logs']['task_logs_separated'] = False
if MAIN_CONFIG['logs']['task_logs_separated'] and 'logs_life_hours' not in MAIN_CONFIG['logs']:
    MAIN_CONFIG['logs']['logs_life_hours'] = 7*24
if 'success_tasks_life_hours' not in MAIN_CONFIG['logs']:
    MAIN_CONFIG['logs']['success_tasks_life_hours'] = 7*24

DEFAULT_LOG_LEVEL = 'INFO'

if 'dir' in MAIN_CONFIG['logs'] and MAIN_CONFIG['logs']['dir'] is not None:

    log_conf = MAIN_CONFIG['logs']
    max_bytes = log_conf['maxBytes'] if 'maxBytes' in log_conf and log_conf['maxBytes'] else 1024*1024*100
    backupCount = log_conf['backupCount'] if 'backupCount' in log_conf and log_conf['backupCount'] else 10
    level = log_conf['level'] if 'level' in log_conf and log_conf['level'] else DEFAULT_LOG_LEVEL
    sentry_enabled = log_conf.get('sentry_enabled', False)

    handlers = ['console', 'common_errors', 'common']

    if sentry_enabled:
        handlers.append('sentry')

    if 'mail_admins' in log_conf and log_conf['mail_admins']:
        handlers.append('mail_admins')

    TASKS_LOGGING = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'default': {
                'format': '%(asctime)s - %(processName)s - %(levelname)s - %(message)s'
            }
        },
        'handlers': {
            'sentry': {
                'level': 'ERROR',
                'class': 'raven.contrib.django.raven_compat.handlers.SentryHandler',
            },
            'mail_admins': {
                'level': 'ERROR',
                'class': 'django.utils.log.AdminEmailHandler',
            },
            'console': {
                'level': level,
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'stream': sys.stdout
            },
            'common_errors': {
                'level': 'ERROR',
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': os.path.join(log_conf['dir'], 'botmanager_errors.log'),
                'maxBytes': max_bytes,
                'formatter': 'default',
                'encoding': 'utf-8',
                'backupCount': backupCount,
            },
            'common': {
                'level': level,
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': os.path.join(log_conf['dir'], 'botmanager.log'),
                'maxBytes': max_bytes,
                'formatter': 'default',
                'encoding': 'utf-8',
                'backupCount': backupCount,
            }
        },
        'loggers': {
            'default': {
                'handlers': handlers,
                'level': level,
                'propagate': False
            },
        }
    }
else:
    TASKS_LOGGING = None



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='django-botmanager',
    version='0.1.4',
    description='Async tasks for django',
    author='Dimoha',
    author_email='dimoha@controlstyle.ru',
    url='https://github.com/dimoha/django-botmanager',
    install_requires=[
        "setproctitle==1.1.10",
        "jsonfield==1.0.3"
    ],
    packages=find_packages(),
)


# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2017-02-17 01:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('botmanager', '0005_task_parent'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='priority',
            field=models.SmallIntegerField(default=0),
        ),
    ]
